package org.fictus

import org.fictus.internal.Method

trait Behavior {
    def handle: Any
}
case class Returns(returnValue: Any) extends Behavior {
    override def handle = returnValue
    override val toString = " --> " + returnValue
}
case class Throws(t: Throwable) extends Behavior {
    override def handle = throw new ExceptionWrapper(t)
    override val toString = " throws " + t.getClass.getSimpleName // + "(" + e.getMessage + ")"
}
case class AnyTimes(behavior: Behavior) extends Behavior {
    override def handle = behavior.handle
    override val toString = behavior + " anyTimes"
}
case class Times(n: Int, behavior: Behavior) extends Behavior {
    override def handle = behavior.handle
    override val toString = behavior + " " + n + ".times"
}
case class Nop() extends Behavior {
    override def handle = ()
    override val toString = ""
}

object Expectation {
    def apply[A](mock: Mock, method: Method, matchers: List[Matcher], args: List[AnyRef], hack: String = "")(implicit context: MockContext): Expectation[A] = {
        val parameterTypes = method.parameterTypes

        context.logging.printSystemError("MATCHERS=" + matchers)

        if (matchers.size > parameterTypes.size)
            throw new MatcherError("You have specified more matchers than arguments... only one matcher per argument will work.")
        if ((!args.isEmpty && !matchers.isEmpty) && (matchers.size != parameterTypes.size))
            throw new MatcherError("If you provide one matcher, you must provide matchers for all of the arguments! Sorry :(")

        val validMatchers = matchers match {
            case Nil => args.map(Equal(_))
            case _   => matchers
        }

        new Expectation(mock, method, validMatchers, args)
    }
}
case class Expectation[A] private[fictus] (mock: Mock, method: Method, matchers: List[Matcher], args: List[AnyRef])(implicit context: MockContext) {

    def returns[B <: A](arg: B) = context.expectations.add(this, Returns(arg))
    def -->[B <: A](arg: B) = returns(arg)

    def throws(t: Throwable) = context.expectations.add(this, Throws(t))

    def anyTimes() = context.expectations.convertLastBehaviorToAnyTimes()

    def times(n: Int) =
        if (n < 2) throw new PebkacError("You may not specify a times < 2 !")
        else context.expectations.convertLastBehaviorToNTimes(n)

    val returnsUnit = method.returnType == classOf[Unit]

    val methodName = mock + "." + method.name + matchers.map(_.expectedMsg).mkString("(", ", ", ")")
    override val toString = methodName
}